@@ -1,12 +1,16 @@
-from functools import lru_cache
+from typing import Tuple, List
 
 import pandas as pd
 import os
 import pickle
 import numpy as np
 import json
-from copy import deepcopy
-from pymatgen import Composition
+from functools import lru_cache, wraps
+
+from pymatgen.entries.computed_entries import ComputedStructureEntry
+from pymongo import MongoClient
+from scipy.interpolate import interp1d
+from pymatgen.core.composition import Composition
 from pymatgen.analysis.substrate_analyzer import SubstrateAnalyzer
 from matminer.featurizers.base import MultipleFeaturizer
 from matminer.featurizers.composition import (
@@ -22,56 +26,97 @@ from matminer.featurizers.structure import (
     StructureComposition,
     MaximumPackingEfficiency,
 )
-from pymatgen.entries.computed_entries import ComputedStructureEntry
 from sklearn.preprocessing import StandardScaler
 from sklearn.metrics import pairwise_distances
 from sklearn.linear_model import LinearRegression
 
 from joblib import Parallel, delayed
 
-from piro.data import ST, H
-from piro import RXN_FILES
+from piro.data import ST
+from piro.settings import settings, CacheType
 
 
-@lru_cache()
-def get_composition(entry: ComputedStructureEntry) -> Composition:
-    return entry.structure.composition
+def get_v(composition_as_dict: dict, elts: Tuple[str]) -> np.array:
+    return np.array([composition_as_dict[el] for el in elts])
 
 
-@lru_cache()
-def get_reduced_formula(entry: ComputedStructureEntry) -> str:
-    return get_composition(entry).reduced_formula
+def get_epitaxies(precursor_library: List[ComputedStructureEntry], target: ComputedStructureEntry):
+    if settings.cache_type == CacheType.MONGO_CACHE:
+        precursor_set = set([s.entry_id for s in precursor_library])
+        epitaxies = {}
+        for e in mongo_results_generator(target, 'epitaxies'):
+            for material_id in e["material_ids"]:
+                if material_id in precursor_set:
+                    # If "min_epitaxy" is not in the cached entry, most likely the epitaxy calculation failed.
+                    # Assign the high value so they will get discounted.
+                    epitaxies[material_id] = float(e.get("min_epitaxy", 1000000.0))
+
+    else:
+        precursor_structures = [s.structure for s in precursor_library]
+        target_structure = target.structure
+        results = epitaxy(precursor_structures, target_structure)
+        epitaxies = dict(
+            zip([i.entry_id for i in precursor_library], results)
+        )
+        print("Epitaxies ready")
+
+    return epitaxies
 
 
-@lru_cache()
-def get_fractional_composition(entry: ComputedStructureEntry) -> Composition:
-    return get_composition(entry).fractional_composition
+def get_similarities(precursor_library: List[ComputedStructureEntry], target: ComputedStructureEntry):
+    if settings.cache_type == CacheType.MONGO_CACHE:
+        precursor_set = set([s.entry_id for s in precursor_library])
+        similarities = {}
+        for s in mongo_results_generator(target, 'similarity'):
+            if "similarity" not in s.keys():
+                print(f"Missing 'similarity' at:\n {s}")
+                continue
+            for material_id in s["material_ids"]:
+                if material_id in precursor_set:
+                    similarities[material_id] = s["similarity"]
+
+    else:
+        precursor_structures = [s.structure for s in precursor_library]
+        target_structure = target.structure
+        results = similarity(precursor_structures, target_structure)
+        similarities = dict(zip([i.entry_id for i in precursor_library], results))
+        print("similarity matrix ready")
+
+    return similarities
 
 
-@lru_cache()
-def get_composition_as_dict(composition: Composition) -> dict:
-    return composition.as_dict()
+def mongo_results_generator(target: ComputedStructureEntry, collection_name: str):
+    with MongoClient(settings.mongodb_uri) as mongo_client:
+        for result in mongo_client.piro[collection_name].find({"material_ids": target.entry_id}):
+            yield result
 
 
-def get_v(c, elts):
-    c_dict = get_composition_as_dict(c)
-    return np.array([c_dict[el] for el in elts])
+def through_cache_if_enabled(func):
+    @wraps(func)
+    def wrapper(_parents, target):
+        if settings.cache_type == CacheType.FILE_CACHE:
+            return through_cache(_parents, target, func)
+        else:
+            return func(_parents, target)
+    return wrapper
 
 
-def through_cache(_parents, target, type="epitaxy"):
-    cache_path = os.path.join(RXN_FILES, "_" + type + "_cache.json")
+def through_cache(_parents, target, func):
+
+    if not os.path.exists(settings.rxn_files):
+        os.mkdir(settings.rxn_files)
+
+    cache_path = os.path.join(settings.rxn_files, "_" + func.__name__ + "_cache.json")
     if os.path.isfile(cache_path):
-        with open(os.path.join(RXN_FILES, "_" + type + "_cache.json"), "r") as f:
+        with open(cache_path, "r") as f:
             db = json.load(f)
     else:
         db = {}
     ordered_pairs = ["_".join(sorted([target.entry_id, i.entry_id])) for i in _parents]
     indices_compt = [i for i in range(len(ordered_pairs)) if ordered_pairs[i] not in db]
 
-    if type == "epitaxy" and indices_compt:
-        _res = epitaxy(np.array(_parents)[indices_compt].tolist(), target)
-    elif type == "similarity" and indices_compt:
-        _res = similarity(np.array(_parents)[indices_compt].tolist(), target)
+    if indices_compt:
+        _res = func(np.array(_parents)[indices_compt].tolist(), target)
     else:
         _res = []
     results = []
@@ -82,23 +127,27 @@ def through_cache(_parents, target, type="epitaxy"):
             results.append(_res[indices_compt.index(ordered_pairs.index(i))])
             db[i] = _res[indices_compt.index(ordered_pairs.index(i))]
 
-    with open(os.path.join(RXN_FILES, "_" + type + "_cache.json"), "w") as f:
+    with open(cache_path, "w") as f:
         json.dump(db, f)
     return results
 
 
+@through_cache_if_enabled
 def epitaxy(_parents, target):
     def _func(s, target):
         sa = SubstrateAnalyzer(film_max_miller=2, substrate_max_miller=2)
-        gen = list(sa.calculate(s, target))
+        gen = [g for g in sa.calculate(s, target) if g]
         if gen:
-            return min([e["match_area"] for e in gen])
+            return min([e.match_area for e in gen])
         else:
             return 1000000.0
 
+    # Uncomment to debug SA issues
+    # return [_func(s, target) for s in _parents]
     return Parallel(n_jobs=-1, verbose=1)(delayed(_func)(s, target) for s in _parents)
 
 
+@through_cache_if_enabled
 def similarity(_parents, target):
     featurizer = MultipleFeaturizer(
         [
@@ -114,6 +163,14 @@ def similarity(_parents, target):
         ]
     )
 
+    # HACK celery doesn't work with multiprocessing (used by matminer)
+    try:
+        from celery import current_task
+        if current_task:
+            featurizer.set_n_jobs(1)
+    except ImportError:
+        pass
+
     x_target = pd.DataFrame.from_records(
         [featurizer.featurize(target)], columns=featurizer.feature_labels()
     )
@@ -127,9 +184,9 @@ def similarity(_parents, target):
     x_target = x_target.reindex(sorted(x_target.columns), axis=1)
     x_parent = x_parent.reindex(sorted(x_parent.columns), axis=1)
 
-    with open(os.path.join(RXN_FILES, "scaler2.pickle"), "rb") as f:
+    with open(os.path.join(settings.rxn_files, "scaler2.pickle"), "rb") as f:
         scaler = pickle.load(f)
-    with open(os.path.join(RXN_FILES, "quantiles.pickle"), "rb") as f:
+    with open(os.path.join(settings.rxn_files, "quantiles.pickle"), "rb") as f:
         quantiles = pickle.load(f)
 
     X = scaler.transform(x_parent.append(x_target))
@@ -144,36 +201,12 @@ def similarity(_parents, target):
     return _res
 
 
-def update_gases(entries, T, P=1, copy=False):
-    """
-    Modify entry objects corresponding to gas phases to account for enthalpy and entropy
-    changes wrt. T and standard pressure.
-    Args:
-        entries (list): List of pymatgen Entry objects.
-        T (float): temperature in K.
-        P (float): pressure dict or float in atm
-        copy (bool): deepcopy entries or update in place.
-    Returns:
-        list of updated entries
-    """
-    if copy:
-        _entries = deepcopy(entries)
-    else:
-        _entries = entries
-    for e in _entries:
-        c = e.composition.reduced_formula
-        if c in ST:
-            if isinstance(P, dict):
-                pp = P.get(c, 1.0)
-            else:
-                pp = P
-            e.data["formation_energy_per_atom"] = (
-                H.get(c, 0.0)
-                - ST[c][T]
-                + 8.6173324e-5 * T * np.log(pp) / Composition(c).num_atoms
-            )
-            e.data["enthalpy"] = H.get(e.composition.reduced_formula, 0.0)
-    return _entries
+@lru_cache()
+def get_ST(c, T):
+    f = interp1d(np.fromiter(ST[c].keys(), dtype=float),
+                 np.fromiter(ST[c].values(), dtype=float),
+                 bounds_error=True)
+    return f(T)
 
 
 def recompute_flatd(source="camd/shared-data/oqmd1.2_icsd_featurized_clean_v2.pickle"):
